//------------------------------------------------------------------------------------
// Copyright (c)2004 Voipac

#include <c8051f300.h>                       // SFR declarations
#include <stdio.h>

#include "ps2.h"

//------------------------------------------------------------------------------------
// global constants

//#define I2C_WRITE			0x00				// i2c WRITE command
#define I2C_READ			0x01				// i2c READ command

#define I2C_KEYB_ADDRESS	0xF0				// keyboard device slave addresses (0x78)
#define I2C_MOUSE_ADDRESS	0xF2				// mouse device slave addresses (0x79)

// status vector definition (top 4 bits only)
#define	I2C_SRADD			0x20				// (SR) slave address received
#define	I2C_SRSTO			0x10				// (SR) STOP detected, or lost arbitration
#define	I2C_SRDB			0x00				// (SR) data byte received, or lost arbitration
#define	I2C_STDB			0x40				// (ST) data byte transmitted
#define	I2C_STSTO			0x50				// (ST) STOP detected

sbit  PS2_IRQ = P0^6;

//-----------------------------------------------------------------------------------
// functions

#define KEYB_WBUF_SIZE		8
#define MOUSE_WBUF_SIZE		8

// keyboard definitions
static uchar keyb_wbuf [KEYB_WBUF_SIZE];
static uchar keyb_win, keyb_wout;

// mouse definitions
static uchar mouse_wbuf [MOUSE_WBUF_SIZE];
static uchar mouse_win, mouse_wout;

void main (void)
{
	// disable watchdog timer 
	PCA0MD &= ~0x40;
	RSTSRC |= 0x02;

	// set internal oscillator to highest setting
	OSCICN |= 0x03;

	// initialize pin configuration
	XBR0 = 0x00;		// no pins skipped
	XBR1 = 0x04;		// enable i2c pins
	XBR2 = 0x40;		// enable crossbar and weak pull-ups
	P0MDIN = 0xFF;		// all P0 pins digital input
	P0MDOUT = 0x00;		// all P0 pins open-drain output
	P0 = 0xFF;

  	// initialize i2c controller
	SMB0CF = 0;		// set slave i2c mode
	SMB0CF |= 0x80;		// enable

	// initialize ps2 controller
	ps2_init();

	// enable interrupts
	EIE1 |= 1;		// i2c interrupt enable
	EA = 1;			// all interrupts enable

	PS2_IRQ = 1;

	while(1)
	{
		if( PS2_KEYB_HASDATA || PS2_MOUSE_HASDATA)
			PS2_IRQ = 0;		// generate falling edge

		if((keyb_win - keyb_wout) != 0)
			ps2_keyb_putchar( keyb_wbuf[(keyb_wout++) & (KEYB_WBUF_SIZE - 1)]);
	
		if((mouse_win - mouse_wout) != 0)
			ps2_mouse_putchar( mouse_wbuf[(mouse_wout++) & (MOUSE_WBUF_SIZE - 1)]);
	}
}

static void i2c_handle_read( uchar scan)
{
	STA = 0;				// clear STA bit
	ACK = 1;				// ACK the received slave address

	SMB0DAT = scan;           // prepare outgoing byte

	// receiver-to-transmitter sequence
	if( !(SMB0DAT & 0x80) )
	{
		STO = 1;			// set STO to force outgoing MSB => '0'
		SI = 0;				// clear i2c interrupt flag

		while( !TXMODE );	// wait for slave to enter transmit mode
		STO = 0;			// clear the STO bit
	}	
}

static uchar i2c_addr;

static uchar i2c_handle_write( uchar c)
{
	switch( i2c_addr)
	{
	case I2C_KEYB_ADDRESS:		// write to keyboard queue
		if (((keyb_win - keyb_wout) & ~(KEYB_WBUF_SIZE-1)) == 0)
		{
			keyb_wbuf [keyb_win++ & (KEYB_WBUF_SIZE-1)] = c;
			return 1;
		}
	break;

	case I2C_MOUSE_ADDRESS:		// write to mouse queue
		if (((mouse_win - mouse_wout) & ~(MOUSE_WBUF_SIZE-1)) == 0)
		{
			mouse_wbuf [mouse_win++ & (MOUSE_WBUF_SIZE-1)] = c;
			return 1;
		}
	break;
	};

	return 0;
}

void i2c_isr(void) interrupt 6
{
	switch (SMB0CN & 0xF0)				// decode the i2c status vector
	{
		// slave receiver: start+address received
		case  I2C_SRADD:
			switch(SMB0DAT)
			{
			case (I2C_KEYB_ADDRESS|I2C_READ):					// keyboard read
				if( PS2_KEYB_HASDATA)
					i2c_handle_read(ps2_keyb_getchar());
				else
					ACK = 0;
				PS2_IRQ = 1;	// clear irq line
			break;

			case (I2C_MOUSE_ADDRESS|I2C_READ):					// mouse read
				if( PS2_MOUSE_HASDATA)
					i2c_handle_read(ps2_mouse_getchar());
				else
					ACK = 0;
				PS2_IRQ = 1;	// clear irq line
			break;

			case I2C_MOUSE_ADDRESS:		// mouse write
			case I2C_KEYB_ADDRESS:		// keyboard write
				i2c_addr = SMB0DAT;
				STA = 0;				// clear STA bit
				ACK = 1;				// ACK the received slave address
			break;

			default:
				ACK = 0;		// NACK received address
			};
		break;

		// slave receiver: data received
		case  I2C_SRDB:
			ACK = i2c_handle_write(SMB0DAT);		// ACK/NACK received data
		break;

		// slave receiver: stop received
		case  I2C_SRSTO:
			STO = 0;			// STO must be cleared
		break;
		
		// slave transmitter: data byte transmitted
		case  I2C_STDB:
		break;
		
		// slave transmitter: arbitration lost, stop detected
		case  I2C_STSTO:
		break;
		
		// default: all other cases undefined
		default:
			SMB0CN &= ~0x40;	// reset communication
			SMB0CN |= 0x40;
		break;
	}

	SI = 0;		// clear i2c interrupt flag
}

